#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_pyppeteer
----------------------------------

Tests for `pyppeteer` module.
"""

import asyncio
import math
from pathlib import Path
import sys
import time
import unittest

from syncer import sync

from pyppeteer import launch
from pyppeteer.errors import PageError, ElementHandleError
from pyppeteer.util import install_asyncio, get_free_port
from server import get_application, BASE_HTML


def setUpModule():
    install_asyncio()


class TestPyppeteer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.port = get_free_port()
        time.sleep(0.1)
        cls.app = get_application()
        cls.server = cls.app.listen(cls.port)
        cls.browser = launch(args=['--no-sandbox'])
        cls.page = sync(cls.browser.newPage())

    @classmethod
    def tearDownClass(cls):
        sync(cls.browser.close())
        cls.server.stop()

    def setUp(self):
        self.url = 'http://localhost:{}/'.format(self.port)
        sync(self.page.goto(self.url))

    @sync
    async def test_get(self):
        self.assertEqual(await self.page.title(), 'main')
        self.assertEqual(self.page.url, self.url)
        self.elm = await self.page.querySelector('h1#hello')
        self.assertTrue(self.elm)
        await self.page.goto('about:blank')
        self.assertEqual(self.page.url, 'about:blank')

    @sync
    async def test_get_https(self):
        await self.page.goto('https://example.com/')
        self.assertEqual(self.page.url, 'https://example.com/')

    @sync
    async def test_plain_text_depr(self):
        with self.assertWarns(DeprecationWarning):
            text = await self.page.plainText()
        self.assertEqual(text.split(), ['Hello', 'link1', 'link2'])

    @sync
    async def test_content(self):
        html = await self.page.content()
        self.assertEqual(html.replace('\n', ''), BASE_HTML.replace('\n', ''))

    @sync
    async def test_evaluate(self):
        await self.page.evaluate('window.__injected = 12;')
        injected = await self.page.evaluate('() => window.__injected')
        self.assertEqual(injected, 12)

    @sync
    async def test_evaluate_return_value(self):
        result = await self.page.evaluate('1 + 2')
        self.assertEqual(result, 3)

    @sync
    async def test_evaluate_return_value_with_semicolon(self):
        result = await self.page.evaluate('1 + 5;')
        self.assertEqual(result, 6)

    @sync
    async def test_evaluate_return_value_with_comment(self):
        result = await self.page.evaluate('2 + 5;\n//some comment')
        self.assertEqual(result, 7)

    @sync
    async def test_evaluate_multi_expression(self):
        result = await self.page.evaluate('''
let a = 2;
let b = 3;
a + b
        ''')
        self.assertEqual(result, 5)

    @sync
    async def test_evaluate_error(self):
        with self.assertRaises(ElementHandleError):
            await self.page.evaluate('not.existing.object')

    @sync
    async def test_evaluate_func(self):
        result = await self.page.evaluate('() => 3 * 7')
        self.assertEqual(result, 21)

    @unittest.skip('Promise is not awaited')
    @sync
    async def test_evaluate_func_promise(self):
        result = await self.page.evaluate('() => Promise.resolve(8 * 7)')
        self.assertEqual(result, 56)

    @sync
    async def test_evaluate_func_args(self):
        result = await self.page.evaluate('(a, b) => a * b', 9, 3)
        self.assertEqual(result, 27)

    @sync
    async def test_evaluate_func_complex_object(self):
        obj = {'a': 1, 'b': 'b'}
        result = await self.page.evaluate('(a) => a', obj)
        self.assertEqual(result, obj)

    @sync
    async def test_evaluate_func_return_none(self):
        result = await self.page.evaluate('() => NaN')
        self.assertIs(result, None)

    @sync
    async def test_evaluate_func_return_minus_zero(self):
        result = await self.page.evaluate('() => -0')
        self.assertEqual(result, -0)

    @sync
    async def test_evaluate_func_return_inf(self):
        result = await self.page.evaluate('() => Infinity')
        self.assertEqual(result, math.inf)

    @sync
    async def test_evaluate_func_return_inf_minus(self):
        result = await self.page.evaluate('() => -Infinity')
        self.assertEqual(result, -math.inf)

    @sync
    async def test_evaluate_func_return_undefined(self):
        result = await self.page.evaluate('() => undefined')
        self.assertEqual(result, None)

    @sync
    async def test_evaluate_func_return_null(self):
        result = await self.page.evaluate('() => null')
        self.assertEqual(result, None)

    @sync
    async def test_evaluate_func_error(self):
        with self.assertRaises(ElementHandleError):
            await self.page.evaluate('() => not.existing.object')

    @sync
    async def test_primitive_handle(self):
        handle = await self.page.evaluateHandle('() => 5')
        is_five = await self.page.evaluate('(a) => Object.is(a, 5)', handle)
        self.assertTrue(is_five)

    @sync
    async def test_element(self):
        elm = await self.page.querySelector('h1')
        text = await self.page.evaluate(
            '(element) => element.textContent', elm)
        self.assertEqual('Hello', text)

    @sync
    async def test_elements(self):
        elms = await self.page.querySelectorAll('a')
        self.assertEqual(len(elms), 2)
        elm1 = elms[0]
        elm2 = elms[1]
        func = 'elm => elm.id'
        # 3.5 does not keep node order.
        if sys.version_info >= (3, 6):
            self.assertEqual(await self.page.evaluate(func, elm1), 'link1')
            self.assertEqual(await self.page.evaluate(func, elm2), 'link2')

    @sync
    async def test_elements_eval(self):
        ln = await self.page.querySelectorAllEval('a', 'nodes => nodes.length')
        self.assertEqual(ln, 2)

    @sync
    async def test_element_inner_html(self):
        elm = await self.page.querySelector('h1')
        text = await self.page.evaluate('(element) => element.innerHTML', elm)
        self.assertEqual('Hello', text)

    @sync
    async def test_element_outer_html(self):
        elm = await self.page.querySelector('h1')
        text = await self.page.evaluate('(element) => element.outerHTML', elm)
        self.assertEqual('<h1 id="hello">Hello</h1>', text)

    @sync
    async def test_element_attr(self):
        _id = await self.page.querySelectorEval('h1', ('(elm) => elm.id'))
        self.assertEqual('hello', _id)

    @sync
    async def test_click(self):
        await self.page.click('#link1')
        await asyncio.sleep(0.05)
        await self.page.waitForSelector('h1#link1')
        self.assertEqual(await self.page.title(), 'link1')
        elm = await self.page.querySelector('h1#link1')
        self.assertTrue(elm)

    @sync
    async def test_tap(self):
        await self.page.tap('#link1')
        await asyncio.sleep(0.05)
        await self.page.waitForSelector('h1#link1')
        self.assertEqual(self.page.url, self.url + '1')
        self.assertEqual(await self.page.title(), 'link1')

    @sync
    async def test_wait_for_timeout(self):
        await self.page.click('#link1')
        await self.page.waitFor(0.1)
        self.assertEqual(await self.page.title(), 'link1')

    @sync
    async def test_wait_for_function(self):
        await self.page.evaluate(
            '() => {'
            '  setTimeout(() => {'
            '    document.body.innerHTML = "<section>a</section>"'
            '  }, 200)'
            '}'
        )
        await asyncio.sleep(0.05)
        await self.page.waitForFunction(
            '() => !!document.querySelector("section")'
        )
        self.assertIsNotNone(await self.page.querySelector('section'))

    @sync
    async def test_wait_for_selector(self):
        await self.page.evaluate(
            '() => {'
            '  setTimeout(() => {'
            '    document.body.innerHTML = "<section>a</section>"'
            '  }, 200)'
            '}'
        )
        await asyncio.sleep(0.05)
        await self.page.waitForSelector('section')
        self.assertIsNotNone(await self.page.querySelector('section'))

    @sync
    async def test_elm_click(self):
        btn1 = await self.page.querySelector('#link1')
        self.assertTrue(btn1)
        await btn1.click()
        await asyncio.sleep(0.05)
        await self.page.waitForSelector('h1#link1')
        self.assertEqual(await self.page.title(), 'link1')

    @sync
    async def test_elm_tap(self):
        btn1 = await self.page.querySelector('#link1')
        self.assertTrue(btn1)
        await btn1.tap()
        await asyncio.sleep(0.05)
        await self.page.waitForSelector('h1#link1')
        self.assertEqual(await self.page.title(), 'link1')

    @sync
    async def test_back_forward(self):
        await self.page.click('#link1')
        await self.page.waitForSelector('h1#link1')
        self.assertEqual(await self.page.title(), 'link1')
        await self.page.goBack()
        await self.page.waitForSelector('h1#hello')
        self.assertEqual(await self.page.title(), 'main')
        elm = await self.page.querySelector('h1#hello')
        self.assertTrue(elm)
        await self.page.goForward()
        await self.page.waitForSelector('h1#link1')
        self.assertEqual(await self.page.title(), 'link1')
        btn2 = await self.page.querySelector('#link1')
        self.assertTrue(btn2)

    @sync
    async def test_cookies(self):
        cookies = await self.page.cookies()
        self.assertEqual(cookies, [])
        await self.page.evaluate(
            'document.cookie = "username=John Doe"'
        )
        cookies = await self.page.cookies()
        self.assertEqual(cookies, [{
            'name': 'username',
            'value': 'John Doe',
            'domain': 'localhost',
            'path': '/',
            'expires': -1,
            'size': 16,
            'httpOnly': False,
            'secure': False,
            'session': True,
        }])
        await self.page.setCookie({'name': 'password', 'value': '123456'})
        cookies = await self.page.evaluate(
            '() => document.cookie'
        )
        self.assertEqual(cookies, 'username=John Doe; password=123456')
        cookies = await self.page.cookies()
        self.assertEqual(cookies, [{
            'name': 'password',
            'value': '123456',
            'domain': 'localhost',
            'path': '/',
            'expires': -1,
            'size': 14,
            'httpOnly': False,
            'secure': False,
            'session': True,
        }, {
            'name': 'username',
            'value': 'John Doe',
            'domain': 'localhost',
            'path': '/',
            'expires': -1,
            'size': 16,
            'httpOnly': False,
            'secure': False,
            'session': True,
        }])
        await self.page.deleteCookie({'name': 'username'})
        cookies = await self.page.evaluate(
            '() => document.cookie'
        )
        self.assertEqual(cookies, 'password=123456')
        cookies = await self.page.cookies()
        self.assertEqual(cookies, [{
            'name': 'password',
            'value': '123456',
            'domain': 'localhost',
            'path': '/',
            'expires': -1,
            'size': 14,
            'httpOnly': False,
            'secure': False,
            'session': True,
        }])

    @sync
    async def test_redirect(self):
        await self.page.goto(self.url + 'redirect1')
        await self.page.waitForSelector('h1#red2')
        text = await self.page.evaluate('() => document.body.innerText')
        self.assertEqual(text, 'redirect2')

    @sync
    async def test_script_tag_error(self):
        await self.page.goto(self.url + 'empty')
        with self.assertRaises(AttributeError):
            await self.page.addScriptTag('/static/injectedfile.js')

    @sync
    async def test_script_tag_url(self):
        await self.page.goto(self.url + 'empty')
        await self.page.addScriptTag(url='/static/injectedfile.js')
        self.assertEqual(await self.page.evaluate('() => window.__injected'), 42)  # noqa: E501

    @sync
    async def test_script_tag_path(self):
        curdir = Path(__file__).parent
        path = str(curdir / 'static' / 'injectedfile.js')
        await self.page.goto(self.url + 'empty')
        await self.page.addScriptTag(path=path)
        self.assertEqual(await self.page.evaluate('() => window.__injected'), 42)  # noqa: E501

    @sync
    async def test_script_tag_content(self):
        await self.page.goto(self.url + 'empty')
        await self.page.addScriptTag(content='window.__injected = 35;')
        self.assertEqual(await self.page.evaluate('() => window.__injected'), 35)  # noqa: E501

    @sync
    async def test_style_tag_error(self):
        await self.page.goto(self.url + 'empty')
        with self.assertRaises(AttributeError):
            await self.page.addStyleTag('/static/injectedstyle.css')

    async def get_bgcolor(self):
        return await self.page.evaluate('() => window.getComputedStyle(document.querySelector("body")).getPropertyValue("background-color")')  # noqa: E501

    @sync
    async def test_style_tag_url(self):
        await self.page.goto(self.url + 'empty')
        self.assertEqual(await self.get_bgcolor(), 'rgba(0, 0, 0, 0)')
        await self.page.addStyleTag(url='/static/injectedstyle.css')
        self.assertEqual(await self.get_bgcolor(), 'rgb(255, 0, 0)')

    @sync
    async def test_style_tag_path(self):
        curdir = Path(__file__).parent
        path = str(curdir / 'static' / 'injectedstyle.css')
        await self.page.goto(self.url + 'empty')
        self.assertEqual(await self.get_bgcolor(), 'rgba(0, 0, 0, 0)')
        await self.page.addStyleTag(path=path)
        self.assertEqual(await self.get_bgcolor(), 'rgb(255, 0, 0)')

    @sync
    async def test_style_tag_content(self):
        await self.page.goto(self.url + 'empty')
        self.assertEqual(await self.get_bgcolor(), 'rgba(0, 0, 0, 0)')
        await self.page.addStyleTag(content=' body {background-color: green;}')
        self.assertEqual(await self.get_bgcolor(), 'rgb(0, 128, 0)')


class TestPage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.port = get_free_port()
        cls.url = 'http://localhost:{}/'.format(cls.port)
        cls.app = get_application()
        time.sleep(0.1)
        cls.server = cls.app.listen(cls.port)
        cls.browser = launch(args=['--no-sandbox'])

    def setUp(self):
        self.page = sync(self.browser.newPage())
        sync(self.page.goto(self.url))

    def tearDown(self):
        sync(self.page.goto('about:blank'))

    @classmethod
    def tearDownClass(cls):
        sync(cls.browser.close())
        cls.server.stop()

    @sync
    async def test_close_page(self):
        await self.page.close()
        self.page = await self.browser.newPage()

    @sync
    async def test_alert(self):
        def dialog_test(dialog):
            self.assertEqual(dialog.type, 'alert')
            self.assertEqual(dialog.defaultValue(), '')
            self.assertEqual(dialog.message(), 'yo')
            asyncio.ensure_future(dialog.accept())
        self.page.on('dialog', dialog_test)
        await self.page.evaluate('() => alert("yo")')

    @sync
    async def test_prompt(self):
        def dialog_test(dialog):
            self.assertEqual(dialog.type, 'prompt')
            self.assertEqual(dialog.defaultValue(), 'yes.')
            self.assertEqual(dialog.message(), 'question?')
            asyncio.ensure_future(dialog.accept('answer!'))
        self.page.on('dialog', dialog_test)
        answer = await self.page.evaluate('() => prompt("question?", "yes.")')
        self.assertEqual(answer, 'answer!')

    @sync
    async def test_prompt_dismiss(self):
        def dismiss_test(dialog, *args):
            asyncio.ensure_future(dialog.dismiss())
        self.page.on('dialog', dismiss_test)
        result = await self.page.evaluate('() => prompt("question?", "yes.")')
        self.assertIsNone(result)

    @sync
    async def test_user_agent(self):
        self.assertIn('Mozilla', await self.page.evaluate(
            '() => navigator.userAgent'))
        await self.page.setUserAgent('foobar')
        await self.page.goto(self.url)
        self.assertEqual('foobar', await self.page.evaluate(
            '() => navigator.userAgent'))

    @sync
    async def test_viewport(self):
        await self.page.setViewport(dict(
            width=480,
            height=640,
            deviceScaleFactor=3,
            isMobile=True,
            hasTouch=True,
            isLandscape=True,
        ))

    @sync
    async def test_emulate(self):
        await self.page.emulate(dict(
            userAgent='test',
            viewport=dict(
                width=480,
                height=640,
                deviceScaleFactor=3,
                isMobile=True,
                hasTouch=True,
                isLandscape=True,
            ),
        ))

    @sync
    async def test_inject_file(self):  # deprecated
        tmp_file = Path('tmp.js')
        with tmp_file.open('w') as f:
            f.write('''
() => document.body.appendChild(document.createElement("section"))
            '''.strip())
        with self.assertWarns(DeprecationWarning):
            await self.page.injectFile(str(tmp_file))
        await self.page.waitForSelector('section')
        self.assertIsNotNone(await self.page.J('section'))
        tmp_file.unlink()

    @sync
    async def test_tracing(self):
        outfile = Path(__file__).parent / 'trace.json'
        if outfile.is_file():
            outfile.unlink()
        await self.page.tracing.start({
            'path': str(outfile)
        })
        await self.page.goto(self.url)
        await self.page.tracing.stop()
        self.assertTrue(outfile.is_file())

    @unittest.skip('This test fails')
    @sync
    async def test_interception_enable(self):
        await self.page.setRequestInterceptionEnabled(True)
        await self.page.goto(self.url)

    @sync
    async def test_auth(self):
        response = await self.page.goto(self.url + 'auth')
        self.assertEqual(response.status, 401)
        await self.page.authenticate({'username': 'user', 'password': 'pass'})
        response = await self.page.goto(self.url + 'auth')
        self.assertEqual(response.status, 200)

    @sync
    async def test_metrics(self):
        await self.page.goto('about:blank')
        metrics = await self.page.getMetrics()
        metrics_to_check = set([
            'Timestamp',
            'Documents',
            'Frames',
            'JSEventListeners',
            'Nodes',
            'LayoutCount',
            'RecalcStyleCount',
            'LayoutDuration',
            'RecalcStyleDuration',
            'ScriptDuration',
            'TaskDuration',
            'JSHeapUsedSize',
            'JSHeapTotalSize',
        ])
        for name, value in metrics.items():
            self.assertTrue(name in metrics_to_check)
            self.assertTrue(value >= 0)
            metrics_to_check.remove(name)
        self.assertEqual(len(metrics_to_check), 0)

    @sync
    async def test_offline_mode(self):
        await self.page.setOfflineMode(True)
        with self.assertRaises(PageError):
            await self.page.goto(self.url)
        await self.page.setOfflineMode(False)
        res = await self.page.reload()
        self.assertEqual(res.status, 304)

    @sync
    async def test_no_await_check_just_call(self):
        await self.page.setExtraHTTPHeaders({'a': 'b'})
        await self.page.setContent('')
        await self.page.setJavaScriptEnabled(True)
        await self.page.emulateMedia()
        await self.page.evaluateOnNewDocument('() => 1 + 2')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import base64
import functools
import os
from typing import Any, Callable

from tornado import web


BASE_HTML = '''
<html>
<head><title>main</title></head>
<body>
<h1 id="hello">Hello</h1>
<a id="link1" href="./1">link1</a>
<a id="link2" href="./2">link2</a>
</body>
</html>
'''


class MainHandler(web.RequestHandler):
    def get(self) -> None:
        self.write(BASE_HTML)


class EmptyHandler(web.RequestHandler):
    def get(self) -> None:
        self.write('')


class LinkHandler1(web.RequestHandler):
    def get(self) -> None:
        self.write('''
<head><title>link1</title></head>
<h1 id="link1">Link1</h1>
<a id="back1" href="./">back1</a>
        ''')


class RedirectHandler1(web.RequestHandler):
    def get(self) -> None:
        self.redirect('/redirect2')


class RedirectHandler2(web.RequestHandler):
    def get(self) -> None:
        self.write('<h1 id="red2">redirect2</h1>')


def auth_api(username: str, password: str) -> bool:
    if username == 'user' and password == 'pass':
        return True
    else:
        return False


def basic_auth(auth: Callable[[str, str], bool]) -> Callable:
    def decore(f: Callable) -> Callable:
        def _request_auth(handler: Any) -> None:
            handler.set_header('WWW-Authenticate', 'Basic realm=JSL')
            handler.set_status(401)
            handler.finish()

        @functools.wraps(f)
        def new_f(*args: Any) -> None:
            handler = args[0]

            auth_header = handler.request.headers.get('Authorization')
            if auth_header is None:
                return _request_auth(handler)
            if not auth_header.startswith('Basic '):
                return _request_auth(handler)

            auth_decoded = base64.b64decode(auth_header[6:])
            username, password = auth_decoded.decode('utf-8').split(':', 2)

            if (auth(username, password)):
                f(*args)
            else:
                _request_auth(handler)

        return new_f
    return decore


class AuthHandler(web.RequestHandler):
    @basic_auth(auth_api)
    def get(self) -> None:
        self.write('ok')


def get_application() -> web.Application:
    static_path = os.path.join(os.path.dirname(__file__), 'static')
    return web.Application([
        ('/', MainHandler),
        ('/1', LinkHandler1),
        ('/redirect1', RedirectHandler1),
        ('/redirect2', RedirectHandler2),
        ('/auth', AuthHandler),
        ('/empty', EmptyHandler),
        ('/static', web.StaticFileHandler, dict(path=static_path)),
    ], logging='error', static_path=static_path)


if __name__ == '__main__':
    import asyncio
    from pyppeteer.util import install_asyncio
    install_asyncio()
    app = get_application()
    app.listen(9000)
    asyncio.get_event_loop().run_forever()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Keyboard and Mouse module."""

import asyncio
from typing import Any, Dict, TYPE_CHECKING

from pyppeteer.connection import Session

if TYPE_CHECKING:
    from typing import Set  # noqa: F401


class Keyboard(object):
    """Keyboard class."""

    def __init__(self, client: Session) -> None:
        """Make new keyboard object."""
        self._client = client
        self._modifiers = 0
        self._pressedKeys: Set[str] = set()

    async def down(self, key: str, options: dict = None, **kwargs: Any
                   ) -> None:
        """Press down key."""
        options = options or dict()
        options.update(kwargs)
        text = options.get('text')
        if text is None:
            text = key if len(key) == 1 and not self._modifiers & ~8 else ''
        self._pressedKeys.add(key)
        self._modifiers |= self._modifierBit(key)
        config = {
            'type': 'rawKeyDown',
            'modifiers': self._modifiers,
            'windowsVirtualKeyCode': keyCodeForKey(key),
            'code': codeForKey(key),
            'key': key,
            'text': text,
        }
        if text:
            config['type'] = 'keyDown'
            config['text'] = text
            config['unmodifiedText'] = text
        if 'autoRepeat' in self._pressedKeys:
            config['autoRepeat'] = True

        await self._client.send('Input.dispatchKeyEvent', config)

    def _modifierBit(self, key: str) -> int:
        if key == 'Alt':
            return 1
        if key == 'Control':
            return 2
        if key == 'Meta':
            return 4
        if key == 'Shift':
            return 8
        return 0

    async def up(self, key: str) -> None:
        """Up pressed key."""
        self._modifiers &= not self._modifierBit(key)
        self._pressedKeys.remove(key)
        await self._client.send('Input.dispatchKeyEvent', {
            'type': 'keyUp',
            'modifiers': self._modifiers,
            'key': key,
            'windowsVirtualKeyCode': keyCodeForKey(key),
            'code': codeForKey(key),
        })

    async def sendCharacter(self, char: str) -> None:
        """Send character."""
        await self._client.send('Input.dispatchKeyEvent', {
            'type': 'char',
            'modifiers': self._modifiers,
            'text': char,
            'key': char,
            'unmodifiedText': char,
        })

    async def type(self, text: str, options: Dict) -> None:
        """Type characters."""
        delay = 0
        if options and options.get('delay'):
            delay = options['delay']
        for char in text:
            await self.press(char, {'text': char, 'delay': delay})
            if delay:
                await asyncio.sleep(delay / 1000)

    async def press(self, key: str, options: Dict) -> None:
        """Press key."""
        await self.down(key, options)
        if options and options.get('delay'):
            await asyncio.sleep(options['delay'] / 1000)
        await self.up(key)


class Mouse(object):
    """Mouse class."""

    def __init__(self, client: Session, keyboard: Keyboard) -> None:
        """Make new mouse object."""
        self._client = client
        self._keyboard = keyboard
        self._x = 0.0
        self._y = 0.0
        self._button = 'none'

    async def move(self, x: float, y: float, options: dict = None,
                   **kwargs: Any) -> None:
        """Move cursor."""
        options = options or dict()
        options.update(kwargs)
        fromX = self._x
        fromY = self._y
        self._x = x
        self._y = y
        steps = options.get('steps', 1)
        for i in range(1, steps + 1):
            x = round(fromX + (self._x - fromX) * (i / steps))
            y = round(fromY + (self._y - fromY) * (i / steps))
            await self._client.send('Input.dispatchMouseEvent', {
                'type': 'mouseMoved',
                'button': self._button,
                'x': x,
                'y': y,
                'modifiers': self._keyboard._modifiers,
            })

    async def click(self, x: float, y: float, options: dict = None,
                    **kwargs: Any) -> None:
        """Click button at (x, y)."""
        if options is None:
            options = dict()
        options.update(kwargs)
        await self.move(x, y)
        await self.down(options)
        if options and options.get('delay'):
            await asyncio.sleep(options.get('delay', 0))
        await self.up(options)

    async def down(self, options: dict = None, **kwargs: Any) -> None:
        """Press down button."""
        if options is None:
            options = dict()
        options.update(kwargs)
        self._button = options.get('button', 'left')
        await self._client.send('Input.dispatchMouseEvent', {
            'type': 'mousePressed',
            'button': self._button,
            'x': self._x,
            'y': self._y,
            'modifiers': self._keyboard._modifiers,
            'clickCount': options.get('clickCount') or 1,
        })

    async def up(self, options: dict = None, **kwargs: Any) -> None:
        """Up pressed button."""
        if options is None:
            options = dict()
        options.update(kwargs)
        self._button = 'none'
        await self._client.send('Input.dispatchMouseEvent', {
            'type': 'mouseReleased',
            'button': options.get('button', 'left'),
            'x': self._x,
            'y': self._y,
            'modifiers': self._keyboard._modifiers,
            'clickCount': options.get('clickCount') or 1,
        })


class Touchscreen(object):
    """Touchscreen class."""

    def __init__(self, client: Session, keyboard: Keyboard) -> None:
        """Make new touchscreen object."""
        self._client = client
        self._keyboard = keyboard

    async def tap(self, x: float, y: float) -> None:
        """Tap (x, y)."""
        touchPoints = [{'x': round(x), 'y': round(y)}]
        await self._client.send('Input.dispatchTouchEvent', {
            'type': 'touchStart',
            'touchPoints': touchPoints,
            'modifiers': self._keyboard._modifiers,
        })
        await self._client.send('Input.dispatchTouchEvent', {
            'type': 'touchEnd',
            'touchPoints': [],
            'modifiers': self._keyboard._modifiers,
        })


keys = {
  'Cancel': {'keyCode': 3, 'code': 'Abort'},
  'Help': {'keyCode': 6, 'code': 'Help'},
  'Backspace': {'keyCode': 8, 'code': 'Backspace'},
  'Tab': {'keyCode': 9, 'code': 'Tab'},
  'Clear': {'keyCode': 12, 'code': ''},
  'Enter': {'keyCode': 13, 'code': 'Enter'},
  'Shift': {'keyCode': 16, 'code': 'ShiftLeft'},
  'Control': {'keyCode': 17, 'code': 'ControlLeft'},
  'Alt': {'keyCode': 18, 'code': 'AltLeft'},
  'Pause': {'keyCode': 19, 'code': 'Pause'},
  'CapsLock': {'keyCode': 20, 'code': 'CapsLock'},
  'Escape': {'keyCode': 27, 'code': 'Escape'},
  'Convert': {'keyCode': 28, 'code': 'Convert'},
  'NonConvert': {'keyCode': 29, 'code': 'NonConvert'},
  'Accept': {'keyCode': 30, 'code': ''},
  'ModeChange': {'keyCode': 31, 'code': ''},
  'PageUp': {'keyCode': 33, 'code': 'PageUp'},
  'PageDown': {'keyCode': 34, 'code': 'PageDown'},
  'End': {'keyCode': 35, 'code': 'End'},
  'Home': {'keyCode': 36, 'code': 'Home'},
  'ArrowLeft': {'keyCode': 37, 'code': 'ArrowLeft'},
  'ArrowUp': {'keyCode': 38, 'code': 'ArrowUp'},
  'ArrowRight': {'keyCode': 39, 'code': 'ArrowRight'},
  'ArrowDown': {'keyCode': 40, 'code': 'ArrowDown'},
  'Select': {'keyCode': 41, 'code': 'Select'},
  'Print': {'keyCode': 42, 'code': ''},
  'Execute': {'keyCode': 43, 'code': 'Open'},
  'PrintScreen': {'keyCode': 44, 'code': 'PrintScreen'},
  'Insert': {'keyCode': 45, 'code': 'Insert'},
  'Delete': {'keyCode': 46, 'code': 'Delete'},
  ')': {'keyCode': 48, 'code': 'Digit0'},
  '!': {'keyCode': 49, 'code': 'Digit1'},
  '@': {'keyCode': 50, 'code': 'Digit2'},
  '#': {'keyCode': 51, 'code': 'Digit3'},
  '$': {'keyCode': 52, 'code': 'Digit4'},
  '%': {'keyCode': 53, 'code': 'Digit5'},
  '^': {'keyCode': 54, 'code': 'Digit6'},
  '&': {'keyCode': 55, 'code': 'Digit7'},
  '*': {'keyCode': 56, 'code': 'Digit8'},
  '(': {'keyCode': 57, 'code': 'Digit9'},
  'Meta': {'keyCode': 91, 'code': 'MetaLeft'},
  'ContextMenu': {'keyCode': 93, 'code': 'ContextMenu'},
  'F1': {'keyCode': 112, 'code': 'F1'},
  'F2': {'keyCode': 113, 'code': 'F2'},
  'F3': {'keyCode': 114, 'code': 'F3'},
  'F4': {'keyCode': 115, 'code': 'F4'},
  'F5': {'keyCode': 116, 'code': 'F5'},
  'F6': {'keyCode': 117, 'code': 'F6'},
  'F7': {'keyCode': 118, 'code': 'F7'},
  'F8': {'keyCode': 119, 'code': 'F8'},
  'F9': {'keyCode': 120, 'code': 'F9'},
  'F10': {'keyCode': 121, 'code': 'F10'},
  'F11': {'keyCode': 122, 'code': 'F11'},
  'F12': {'keyCode': 123, 'code': 'F12'},
  'F13': {'keyCode': 124, 'code': 'F13'},
  'F14': {'keyCode': 125, 'code': 'F14'},
  'F15': {'keyCode': 126, 'code': 'F15'},
  'F16': {'keyCode': 127, 'code': 'F16'},
  'F17': {'keyCode': 128, 'code': 'F17'},
  'F18': {'keyCode': 129, 'code': 'F18'},
  'F19': {'keyCode': 130, 'code': 'F19'},
  'F20': {'keyCode': 131, 'code': 'F20'},
  'F21': {'keyCode': 132, 'code': 'F21'},
  'F22': {'keyCode': 133, 'code': 'F22'},
  'F23': {'keyCode': 134, 'code': 'F23'},
  'F24': {'keyCode': 135, 'code': 'F24'},
  'NumLock': {'keyCode': 144, 'code': 'NumLock'},
  'ScrollLock': {'keyCode': 145, 'code': 'ScrollLock'},
  'AudioVolumeMute': {'keyCode': 173, 'code': 'AudioVolumeMute'},
  'AudioVolumeDown': {'keyCode': 174, 'code': 'AudioVolumeDown'},
  'AudioVolumeUp': {'keyCode': 175, 'code': 'AudioVolumeUp'},
  'MediaTrackNext': {'keyCode': 176, 'code': 'MediaTrackNext'},
  'MediaTrackPrevious': {'keyCode': 177, 'code': 'MediaTrackPrevious'},
  'MediaStop': {'keyCode': 178, 'code': 'MediaStop'},
  'MediaPlayPause': {'keyCode': 179, 'code': 'MediaPlayPause'},
  ';': {'keyCode': 186, 'code': 'Semicolon'},
  ':': {'keyCode': 186, 'code': 'Semicolon'},
  '=': {'keyCode': 187, 'code': 'Equal'},
  '+': {'keyCode': 187, 'code': 'Equal'},
  ',': {'keyCode': 188, 'code': 'Comma'},
  '<': {'keyCode': 188, 'code': 'Comma'},
  '-': {'keyCode': 189, 'code': 'Minus'},
  '_': {'keyCode': 189, 'code': 'Minus'},
  '.': {'keyCode': 190, 'code': 'Period'},
  '>': {'keyCode': 190, 'code': 'Period'},
  '/': {'keyCode': 191, 'code': 'Slash'},
  '?': {'keyCode': 191, 'code': 'Slash'},
  '`': {'keyCode': 192, 'code': 'Backquote'},
  '~': {'keyCode': 192, 'code': 'Backquote'},
  '[': {'keyCode': 219, 'code': 'BracketLeft'},
  '{': {'keyCode': 219, 'code': 'BracketLeft'},
  '\\': {'keyCode': 220, 'code': 'Backslash'},
  '|': {'keyCode': 220, 'code': 'Backslash'},
  ']': {'keyCode': 221, 'code': 'BracketRight'},
  '}': {'keyCode': 221, 'code': 'BracketRight'},
  '\'': {'keyCode': 222, 'code': 'Quote'},
  '"': {'keyCode': 222, 'code': 'Quote'},
  'AltGraph': {'keyCode': 225, 'code': 'AltGraph'},
  'Attn': {'keyCode': 246, 'code': ''},
  'CrSel': {'keyCode': 247, 'code': 'Props'},
  'ExSel': {'keyCode': 248, 'code': ''},
  'EraseEof': {'keyCode': 249, 'code': ''},
  'Play': {'keyCode': 250, 'code': ''},
  'ZoomOut': {'keyCode': 251, 'code': ''},
  '0': {'keyCode': 48, 'code': 'Digit0'},
  '1': {'keyCode': 49, 'code': 'Digit1'},
  '2': {'keyCode': 50, 'code': 'Digit2'},
  '3': {'keyCode': 51, 'code': 'Digit3'},
  '4': {'keyCode': 52, 'code': 'Digit4'},
  '5': {'keyCode': 53, 'code': 'Digit5'},
  '6': {'keyCode': 54, 'code': 'Digit6'},
  '7': {'keyCode': 55, 'code': 'Digit7'},
  '8': {'keyCode': 56, 'code': 'Digit8'},
  '9': {'keyCode': 57, 'code': 'Digit9'},
  'q': {'keyCode': 81, 'code': 'KeyQ'},
  'w': {'keyCode': 87, 'code': 'KeyW'},
  'e': {'keyCode': 69, 'code': 'KeyE'},
  'r': {'keyCode': 82, 'code': 'KeyR'},
  't': {'keyCode': 84, 'code': 'KeyT'},
  'y': {'keyCode': 89, 'code': 'KeyY'},
  'u': {'keyCode': 85, 'code': 'KeyU'},
  'i': {'keyCode': 73, 'code': 'KeyI'},
  'o': {'keyCode': 79, 'code': 'KeyO'},
  'p': {'keyCode': 80, 'code': 'KeyP'},
  'a': {'keyCode': 65, 'code': 'KeyA'},
  's': {'keyCode': 83, 'code': 'KeyS'},
  'd': {'keyCode': 68, 'code': 'KeyD'},
  'f': {'keyCode': 70, 'code': 'KeyF'},
  'g': {'keyCode': 71, 'code': 'KeyG'},
  'h': {'keyCode': 72, 'code': 'KeyH'},
  'j': {'keyCode': 74, 'code': 'KeyJ'},
  'k': {'keyCode': 75, 'code': 'KeyK'},
  'l': {'keyCode': 76, 'code': 'KeyL'},
  'z': {'keyCode': 90, 'code': 'KeyZ'},
  'x': {'keyCode': 88, 'code': 'KeyX'},
  'c': {'keyCode': 67, 'code': 'KeyC'},
  'v': {'keyCode': 86, 'code': 'KeyV'},
  'b': {'keyCode': 66, 'code': 'KeyB'},
  'n': {'keyCode': 78, 'code': 'KeyN'},
  'm': {'keyCode': 77, 'code': 'KeyM'},
  'Q': {'keyCode': 81, 'code': 'KeyQ'},
  'W': {'keyCode': 87, 'code': 'KeyW'},
  'E': {'keyCode': 69, 'code': 'KeyE'},
  'R': {'keyCode': 82, 'code': 'KeyR'},
  'T': {'keyCode': 84, 'code': 'KeyT'},
  'Y': {'keyCode': 89, 'code': 'KeyY'},
  'U': {'keyCode': 85, 'code': 'KeyU'},
  'I': {'keyCode': 73, 'code': 'KeyI'},
  'O': {'keyCode': 79, 'code': 'KeyO'},
  'P': {'keyCode': 80, 'code': 'KeyP'},
  'A': {'keyCode': 65, 'code': 'KeyA'},
  'S': {'keyCode': 83, 'code': 'KeyS'},
  'D': {'keyCode': 68, 'code': 'KeyD'},
  'F': {'keyCode': 70, 'code': 'KeyF'},
  'G': {'keyCode': 71, 'code': 'KeyG'},
  'H': {'keyCode': 72, 'code': 'KeyH'},
  'J': {'keyCode': 74, 'code': 'KeyJ'},
  'K': {'keyCode': 75, 'code': 'KeyK'},
  'L': {'keyCode': 76, 'code': 'KeyL'},
  'Z': {'keyCode': 90, 'code': 'KeyZ'},
  'X': {'keyCode': 88, 'code': 'KeyX'},
  'C': {'keyCode': 67, 'code': 'KeyC'},
  'V': {'keyCode': 86, 'code': 'KeyV'},
  'B': {'keyCode': 66, 'code': 'KeyB'},
  'N': {'keyCode': 78, 'code': 'KeyN'},
  'M': {'keyCode': 77, 'code': 'KeyM'}
}


def keyCodeForKey(key: str) -> int:
    """Get key code for key."""
    if keys.get(key):
        return keys[key]['keyCode']  # type: ignore
    if len(key) == 1:
        return ord(key.upper())
    return 0


def codeForKey(key: str) -> str:
    """Get code for key."""
    if not keys.get(key):
        return ''
    return keys[key]['code']  # type: ignore



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Element handle module."""

import logging
import os.path
from typing import Any, Dict, TYPE_CHECKING

# from pyppeteer import helper
from pyppeteer.connection import Session
from pyppeteer.execution_context import ExecutionContext, JSHandle
from pyppeteer.errors import ElementHandleError

if TYPE_CHECKING:
    from pyppeteer.frame_manager import Frame  # noqa: F401
    # from pyppeteer.page import Page  # noqa: F401


logger = logging.getLogger(__name__)


class ElementHandle(JSHandle):
    """ElementHandle class."""

    def __init__(self, context: ExecutionContext, client: Session,
                 remoteObject: dict, page: Any) -> None:
        """Make new element handle object."""
        super().__init__(context, client, remoteObject)
        self._client = client
        self._remoteObject = remoteObject
        self._page = page
        self._disposed = False

    def asElement(self) -> 'ElementHandle':
        """Return as an element handle."""
        return self

    async def _scrollIntoViewIfNeeded(self) -> None:
        error = await self.executionContext.evaluate(
            '''element => {
                if (!element.ownerDocument.contains(element))
                    return 'Node is detached from document';
                if (element.nodeType !== Node.ELEMENT_NODE)
                    return 'Node is not of type HTMLElement';
                element.scrollIntoViewIfNeeded();
                return false;
            }''', self)
        if error:
            raise ElementHandleError(error)

    async def _visibleCenter(self) -> Dict[str, float]:
        await self._scrollIntoViewIfNeeded()
        box = await self.boundingBox()
        return {
            'x': box['x'] + box['width'] / 2,
            'y': box['y'] + box['height'] / 2,
        }

    async def hover(self) -> None:
        """Move mouse over this element."""
        obj = await self._visibleCenter()
        x = obj.get('x', 0)
        y = obj.get('y', 0)
        await self._page.mouse.move(x, y)

    async def click(self, options: dict = None, **kwargs: Any) -> None:
        """Click this element."""
        obj = await self._visibleCenter()
        x = obj.get('x', 0)
        y = obj.get('y', 0)
        if options is None:
            options = dict()
        options.update(kwargs)
        await self._page.mouse.click(x, y, options)

    async def uploadFile(self, *filePaths: str) -> dict:
        """Upload files."""
        files = [os.path.abspath(p) for p in filePaths]
        objectId = self._remoteObject.get('objectId')
        return await self._client.send(
            'DOM.setFileInputFiles',
            {'objectId': objectId, 'files': files}
        )

    async def tap(self) -> None:
        """Tap this element."""
        center = await self._visibleCenter()
        x = center.get('x', 0)
        y = center.get('y', 0)
        await self._page.touchscreen.tap(x, y)

    async def focus(self) -> None:
        """Focus this element."""
        await self.executionContext.evaluate(
            'element => element.focus()', self)

    async def type(self, text: str, options: Dict = None) -> None:
        """Type characters on this element."""
        if options is None:
            options = {}
        await self.focus()
        await self._page.keyboard.type(text, options)

    async def press(self, key: str, options: Dict = None) -> None:
        """Type key on this element."""
        if options is None:
            options = {}
        await self.focus()
        await self._page.keyboard.press(key, options)

    async def boundingBox(self) -> Dict[str, float]:
        """Return bounding box size of this node."""
        _obj = await self._client.send('DOM.getBoxModel', {
            'objectId': self._remoteObject.get('objectId'),
        })
        model = _obj.get('model')
        if not model:
            raise ElementHandleError('node is detached from document')

        quad = model['border']
        x = min(quad[0], quad[2], quad[4], quad[6])
        y = min(quad[1], quad[3], quad[5], quad[7])
        width = max(quad[0], quad[2], quad[4], quad[6]) - x
        height = max(quad[1], quad[3], quad[5], quad[7]) - y
        return {'x': x, 'y': y, 'width': width, 'height': height}

    async def screenshot(self, options: Dict = None) -> bytes:
        """Take a screenshot of this element."""
        if options is None:
            options = {}
        await self._scrollIntoViewIfNeeded()
        boundingBox = await self.boundingBox()
        opt = {'clip': boundingBox}
        opt.update(options)
        return await self._page.screenshot(opt)
